/**
 * 
 */
package org.processmining.plugins.pnalignanalysis.visualization.performancematrix;

import gnu.trove.list.TIntList;
import gnu.trove.list.TLongList;
import gnu.trove.list.array.TIntArrayList;
import gnu.trove.list.array.TLongArrayList;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.petrinet.replayresult.StepTypes;
import org.processmining.plugins.replayer.replayresult.SyncReplayResult;

/**
 * @author aadrians Nov 19, 2012
 * 
 */
public class PerfMatrixAlignmentProviderBetweenStarts {


	protected Transition[] int2trans = null;
	// [to index 0,..., n from index 0][to index 0, ..., n from index 1]
	protected double[] data = null; // min, max, average, stddev
	protected int[] occ = null;

	public PerfMatrixAlignmentProviderBetweenStarts(PetrinetGraph net, Marking initMarking, XLog log, PNRepResult logReplayResult,
			boolean showInviTrans) {
		// encode transitions
		List<Transition> transitions = new ArrayList<Transition>();
		for (Transition t : net.getTransitions()) {
			if ((t.isInvisible()) && (!showInviTrans)) {
				continue;
			}
			transitions.add(t);
		}
		;

		Collections.sort(transitions, new Comparator<Transition>() {
			public int compare(Transition o1, Transition o2) {
				return o1.getLabel().compareTo(o2.getLabel());
			}

		});

		int2trans = transitions.toArray(new Transition[transitions.size()]);
		Map<Transition, Integer> trans2int = new HashMap<Transition, Integer>(int2trans.length);
		for (int i = 0; i < int2trans.length; i++) {
			trans2int.put(int2trans[i], i);
		}

		// variables to measure averages and standard deviations
		int blockSize = int2trans.length * int2trans.length;
		data = new double[blockSize * 4];
		Arrays.fill(data, Double.NaN);
		occ = new int[blockSize]; // number of occurrence
		Arrays.fill(occ, 0);

		// calculate elapsed time between transitions
		calculateStats(logReplayResult, int2trans, trans2int, log, blockSize);

	}

	protected void calculateStats(final PNRepResult logReplayResult, final Transition[] int2trans,
			final Map<Transition, Integer> trans2int, final XLog log, int blockSize) {
		// calculate stats pairwise
		// utilities objects
		XTimeExtension tx = XTimeExtension.instance();
		TLongList longList = new TLongArrayList(10);
		TIntList transList = new TIntArrayList(10);

		// [from t1 to t1, from t1 to t2, from t2 to t3, ... ][from t2 to t1, ....]
		for (SyncReplayResult srr : logReplayResult) {
			for (int traceIndex : srr.getTraceIndex()) {
				// reset accumulator
				longList.clear();
				transList.clear();
				
				// trace specific information
				Iterator<XEvent> it = log.get(traceIndex).iterator();
				Iterator<Object> nit = srr.getNodeInstance().iterator();
				Iterator<StepTypes> stit = srr.getStepTypes().iterator();

				while (it.hasNext()) { // only need to derive from event
					switch (stit.next()) {
						case L :
							nit.next();
							it.next();
							break;
						case LMGOOD :
							// extract time information
							longList.add(tx.extractTimestamp(it.next()).getTime());
							transList.add(trans2int.get(nit.next()));
							break;
						case LMNOGOOD : 
						case LMREPLACED :
						case LMSWAPPED:
							// deviations other than move model and move log are ignored
							nit.next();
							it.next();
							break;
						case MINVI :
							nit.next();
							break;
						case MREAL :
							longList.add(-Long.MAX_VALUE);
							transList.add(trans2int.get(nit.next()));
							break;
					}
					;
				}
				
				// then, calculate stats iteratively
				// for all FROM trans, boolean indicates used instance of the TO trans
				BitSet usedIndex = new BitSet(longList.size() * longList.size());

				// local analysis
				BitSet localUsedIndex = new BitSet(int2trans.length);
				int upLimit = transList.size()  - 1;
				for (int i = 0; i < upLimit; i++) {
					int fromTransIndex = transList.get(i);
					long fromTime = longList.get(i);

					int j = i + 1;
					localUsedIndex.clear();
					while ((j < transList.size()) && (localUsedIndex.nextClearBit(0) < int2trans.length)) {
						long toTime = longList.get(j);
						int toTransIdx = transList.get(j); 
						j++;

						if ((!localUsedIndex.get(toTransIdx))
								&& (!usedIndex.get((longList.size() * fromTransIndex) + j))) { // if not set
							localUsedIndex.set(toTransIdx); // set
							usedIndex.set((longList.size() * fromTransIndex) + j); // set

							// calculate performance
							if ((fromTime >= 0L) && (toTime >= 0L)) {
								updateMandSVal(fromTransIndex, toTransIdx, data, occ,
										toTime - fromTime, int2trans.length);
							}
						}
					}
					;
				}
			}
		}
		
		// finalize calculation of average and stddev
		for (int i = 0; i < blockSize; i++) {
			int loc = 4 * i;
			data[loc + 3] = Math.sqrt(data[loc + 3] / (occ[i] - 1));
		}
	}


	protected void updateMandSVal(int fromIndex, int toIndex, double[] data, int[] occ, long val, int length) {
		int loc = (length * toIndex) + fromIndex;
		int pos = 4 * loc;
		if (Double.isNaN(data[pos])) {
			data[pos] = val; // min
			data[pos + 1] = val; // max
			data[pos + 2] = val; // mval
			data[pos + 3] = 0; // sval
			occ[loc] = 1; // occurrence
		} else {
			occ[loc]++;
			if (Double.compare(data[pos], val) > 0) {
				data[pos] = val; // min				
			}
			if (Double.compare(data[pos + 1], val) < 0) {
				data[pos + 1] = val; // max
			}
			double oldMVal = data[pos + 2];
			data[pos + 2] += (val - oldMVal) / occ[loc];
			data[pos + 3] += ((val - oldMVal) * (val - data[pos + 2]));
		}
	}

	/**
	 * get only labels of transitions
	 * 
	 * @return
	 */
	public String[] getTransLabel() {
		String[] newStr = new String[int2trans.length];
		for (int i = 0; i < int2trans.length; i++) {
			newStr[i] = int2trans[i].getLabel();
		}
		return newStr;
	}

	/**
	 * Get number of transitions
	 * 
	 * @return
	 */
	public int getNumTrans() {
		return int2trans.length;
	}

	/**
	 * get encoding of transitions
	 * 
	 * @return
	 */
	public Transition[] getInt2Trans() {
		return int2trans;
	}

	/**
	 * 
	 * @param toIndex
	 *            row index of transitions (TO)
	 * @param fromIndex
	 *            column index of transitions (FROM)
	 * @return
	 */
	public double getTimeBetween(int fromIndex, int toIndex, PerfMatrixStats metric) {
		int loc = 4 * ((toIndex * getNumTrans()) + fromIndex);
		switch (metric) {
			case MIN :
				return (this.data[loc]);
			case MAX :
				return (this.data[loc + 1]);
			case AVG :
				return (this.data[loc + 2]);
			case STDDEV :
				return (this.data[loc + 3]);
		}
		return Double.NaN;
	}

	/**
	 * Get frequency i occurred before j
	 * 
	 * @param i
	 * @param j
	 * @return
	 */
	public int getFrequency(int i, int j) {
		return (this.occ[(j * getNumTrans()) + i]);
	}

	/**
	 * get all information
	 * ordered: from 1 to 1, from 1 to 2, ... from 2 to 1, ...
	 * 
	 * @return
	 */
	public double[] getAllInfo() {
		double[] data = new double[getNumTrans() * getNumTrans()];
		Arrays.fill(data, Double.NaN);

		int dataPtr = 0;
		for (int i = 0; i < getNumTrans(); i++) {
			for (int j = 0; j < getNumTrans(); j++) {
				data[dataPtr] = getTimeBetween(i, j, PerfMatrixStats.AVG);
				dataPtr++;
			}
		}
		return data;
	}

	public Map<Transition, Integer> constructMapTrans2Int() {
		Map<Transition, Integer> map = new HashMap<Transition, Integer>(int2trans.length);
		for (int i=0; i < int2trans.length; i++){
			map.put(int2trans[i], i);
		}
		return map;
	}
}
